"""
Handles the merging of two or more model instances.

Adapted from the django-extensions package:
https://github.com/django-extensions/django-extensions
"""

from django.apps import apps
from django.contrib.contenttypes.fields import GenericForeignKey
from django.db import transaction

from uniauth.utils import get_setting


def _get_generic_fields():
    """
    Return a list of all GenericForeignKeys in all models.
    """
    generic_fields = []
    for model in apps.get_models():
        for field_name, field in model.__dict__.items():
            if isinstance(field, GenericForeignKey):
                generic_fields.append(field)
    return generic_fields


@transaction.atomic()
def merge_model_instances(primary_object, alias_objects, field_trace=[]):
    """
    Merge several model instances into one, the `primary_object`.
    Use this function to merge model objects and migrate all of the
    related fields from the alias objects into the primary object.

    Performs recursive merging of related One-to-One fields.
    """
    generic_fields = _get_generic_fields()

    # get related fields
    related_fields = list(
        filter(
            lambda x: x.is_relation is True, primary_object._meta.get_fields()
        )
    )

    many_to_many_fields = list(
        filter(lambda x: x.many_to_many is True, related_fields)
    )

    related_fields = list(
        filter(lambda x: x.many_to_many is False, related_fields)
    )

    # Loop through all alias objects and migrate their references to the
    # primary object
    deleted_objects = []
    deleted_objects_count = 0
    for alias_object in alias_objects:
        # Migrate all foreign key references from alias object to primary
        # object.
        for many_to_many_field in many_to_many_fields:
            alias_varname = many_to_many_field.name
            related_objects = getattr(alias_object, alias_varname, None)
            if related_objects is None:
                continue
            for obj in related_objects.all():
                try:
                    # Handle regular M2M relationships.
                    getattr(alias_object, alias_varname).remove(obj)
                    getattr(primary_object, alias_varname).add(obj)
                except AttributeError:
                    # Handle M2M relationships with a 'through' model.
                    # This does not delete the 'through model.
                    through_model = getattr(
                        alias_object, alias_varname
                    ).through
                    kwargs = {
                        many_to_many_field.m2m_reverse_field_name(): obj,
                        many_to_many_field.m2m_field_name(): alias_object,
                    }
                    through_model_instances = through_model.objects.filter(
                        **kwargs
                    )
                    for instance in through_model_instances:
                        # Re-attach the through model to the primary_object
                        setattr(
                            instance,
                            many_to_many_field.m2m_field_name(),
                            primary_object,
                        )
                        instance.save()

        for related_field in related_fields:
            if related_field.one_to_many:
                alias_varname = related_field.get_accessor_name()
                related_objects = getattr(alias_object, alias_varname, None)
                if related_objects is None:
                    continue
                for obj in related_objects.all():
                    field_name = related_field.field.name
                    setattr(obj, field_name, primary_object)
                    obj.save()
            elif related_field.one_to_one or related_field.many_to_one:
                alias_varname = related_field.name
                related_object = getattr(alias_object, alias_varname, None)
                primary_related_object = getattr(
                    primary_object, alias_varname, None
                )
                if related_object is None:
                    continue
                elif primary_related_object is None:
                    setattr(primary_object, alias_varname, related_object)
                    primary_object.save()
                elif related_field.one_to_one:
                    # Perform recursive merging for one-to-one fields
                    if get_setting("UNIAUTH_PERFORM_RECURSIVE_MERGING"):
                        if related_field in field_trace:
                            continue
                        updated_trace = field_trace + [
                            related_field,
                            related_field.remote_field,
                        ]
                        merge_model_instances(
                            primary_related_object,
                            [related_object],
                            updated_trace,
                        )
                    else:
                        related_object.delete()

        for field in generic_fields:
            filter_kwargs = {}
            filter_kwargs[field.fk_field] = alias_object._get_pk_val()
            filter_kwargs[field.ct_field] = field.get_content_type(
                alias_object
            )
            related_objects = field.model.objects.filter(**filter_kwargs)
            for generic_related_object in related_objects:
                setattr(generic_related_object, field.name, primary_object)
                generic_related_object.save()

        if alias_object.id:
            deleted_objects += [alias_object]
            alias_object.delete()
            deleted_objects_count += 1

    return primary_object, deleted_objects, deleted_objects_count



import json
from contextlib import contextmanager

try:
    from urlparse import parse_qs, urlparse
except ImportError:
    from urllib.parse import parse_qs, urlparse


def assert_urls_equivalent(actual, expected, assert_equal_fn):
    """
    Checks if the the two URLs are funcationally equivalent,
    forwarding to the provided assert_equal_fn if they are not
    """
    actual_path = actual.split("?")[0]
    expected_path = expected.split("?")[0]
    actual_params = parse_qs(urlparse(actual).query)
    expected_params = parse_qs(urlparse(expected).query)
    if actual_path != expected_path or actual_params != expected_params:
        assert_equal_fn(actual, expected)


def pretty_str(list_or_dict):
    """
    Returns a string for the provided list or dict
    that looks good when printed to the console
    """
    return json.dumps(list_or_dict, indent=2)



"""
This command is used to migrate a project previously using
custom User authentication over to Uniauth.

Execution: python manage.py migrate_custom
"""

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.db import transaction

from uniauth.models import LinkedEmail, UserProfile
from uniauth.utils import get_input


class Command(BaseCommand):
    help = "Migrates a project using custom User auhentication to Uniauth."

    @transaction.atomic
    def handle(self, *args, **options):
        message = (
            "This command is intended to migrate projects "
            "previously using custom User authentication to using Uniauth.\n\n"
            "You should only proceed with this command if your project had "
            "users sign up with a username / email address and password. This "
            "command will create UserProfile for each user with a username or "
            "email address, and a non-blank password. A verified LinkedEmail "
            "will also be created if the email field is non-blank.\n\n"
            "Do you still wish to continue?\n\nAnswer [y/n]: "
        )
        answer = get_input(message)

        if answer != "y" and answer != "yes":
            self.stdout.write("\nCanceled.\n")
            return

        self.stdout.write("\nProceeding... ")
        skipped = []
        for user in get_user_model().objects.all():
            # Skip users that already have UserProfiles
            if hasattr(user, "uniauth_profile"):
                continue
            # Skip users lacking a username/email address or password
            if (not user.username and not user.email) or not user.password:
                skipped.append(user.username or user.email or "(none)")
                continue
            # Add the profile + LinkedEmail if email field is non-blank
            profile = UserProfile.objects.create(user=user)
            if user.email:
                LinkedEmail.objects.create(
                    profile=profile, address=user.email, is_verified=True
                )
        self.stdout.write("Done!\n")

        if len(skipped) > 0:
            self.stdout.write(
                "\nThe following users could not be "
                + "migrated: %s\n" % str(skipped)
            )



from random import randint

from django.contrib.auth.models import AnonymousUser, User
from django.test import RequestFactory, TestCase, override_settings

from tests.utils import assert_urls_equivalent, pretty_str
from uniauth.utils import (
    DEFAULT_SETTING_VALUES,
    choose_username,
    decode_pk,
    encode_pk,
    flush_old_tmp_users,
    get_account_username_split,
    get_random_username,
    get_redirect_url,
    get_service_url,
    get_setting,
    is_tmp_user,
)


class ChooseUsernameTests(TestCase):
    """
    Tests the choose_username method in utils.py
    """

    def test_choose_username_unique(self):
        """
        Ensure that all outputted usernames are unique
        """
        emails = [
            "fakeemail@example.com",
            "abcde@fgh.com",
            "abcd@gh.com",
            "abce@gh.com",
            "abce@gh.ijk",
            "repeat@gmail.com",
            "repeat@gmail.com",
            "repeat@gmail.com",
        ]
        usernames = []
        for email in emails:
            username = choose_username(email)
            usernames.append(username)
            User.objects.create(username=username, email=email)
        if len(set(usernames)) != len(emails):
            zip_str = pretty_str(
                [x + ": " + y for x, y in zip(emails, usernames)]
            )
            self.fail(
                "choose_username outputted a duplicate username.\n"
                "Emails vs Usernames: %s" % zip_str
            )


class EncodeDecodePkTests(TestCase):
    """
    Tests the encode_pk and decode_pk methods in utils.py
    """

    def test_encode_decode_pk_reversible(self):
        """
        Ensure that encoding a PK, then decoding it returns
        the original value
        """
        for i in range(50):
            random_pk = randint(1, 10000)
            encoded = encode_pk(random_pk)
            decoded = decode_pk(encoded)
            if str(random_pk) != decoded:
                self.fail(
                    (
                        "Encoding then decoding PK '%d' did not return "
                        "original value.\nEncoded: %s\tDecoded: %s"
                    )
                    % (random_pk, encoded, decoded)
                )


class FlushOldTmpUsersTests(TestCase):
    """
    Tests the flush_old_tmp_users method in utils.py
    """

    def setUp(self):
        from datetime import timedelta

        from django.utils import timezone

        self.tmp0 = User.objects.create(
            username="tmp-0", date_joined=timezone.now() - timedelta(days=0)
        )
        self.tmp1 = User.objects.create(
            username="tmp-1", date_joined=timezone.now() - timedelta(days=1)
        )
        self.tmp2 = User.objects.create(
            username="tmp-2", date_joined=timezone.now() - timedelta(days=2)
        )
        self.tmp3 = User.objects.create(
            username="tmp-3", date_joined=timezone.now() - timedelta(days=3)
        )
        self.real = User.objects.create(
            username="a-real-user",
            date_joined=timezone.now() - timedelta(days=3),
        )

    def test_flush_old_tmp_users_deletes_proper_users(self):
        """
        Ensure the proper users are deleted given the days argument
        """
        num_deleted = flush_old_tmp_users(days=2)
        self.assertEqual(num_deleted, 2)
        remaining_users = User.objects.all()
        self.assertIn(self.tmp0, remaining_users)
        self.assertIn(self.tmp1, remaining_users)
        self.assertNotIn(self.tmp2, remaining_users)
        self.assertNotIn(self.tmp3, remaining_users)
        self.assertIn(self.real, remaining_users)

    def test_flush_old_tmp_users_deletes_proper_users(self):
        """
        Ensure the function uses a default days parameter if not provided
        """
        num_deleted = flush_old_tmp_users()
        self.assertEqual(num_deleted, 3)
        remaining_users = User.objects.all()
        self.assertIn(self.tmp0, remaining_users)
        self.assertNotIn(self.tmp1, remaining_users)
        self.assertNotIn(self.tmp2, remaining_users)
        self.assertNotIn(self.tmp3, remaining_users)
        self.assertIn(self.real, remaining_users)


class GetAccountUsernameSplitTests(TestCase):
    """
    Tests the get_account_username_split method in utils.py
    """

    def test_get_account_username_split_unlinked_account(self):
        """
        Ensure it works for unlinked InstitutionAccount users
        """

        def _run_test(username, expected_result):
            result = get_account_username_split(username)
            self.assertEqual(result[0], expected_result[0])
            self.assertEqual(result[1], expected_result[1])
            self.assertEqual(result[2], expected_result[2])

        _run_test("cas-princeton-netid", ("cas", "princeton", "netid"))
        _run_test("exauth-exinst-exid", ("exauth", "exinst", "exid"))
        _run_test(
            "ab_c-long-dashed-slug-id123",
            ("ab_c", "long-dashed-slug", "id123"),
        )
        _run_test("a-b-c", ("a", "b", "c"))
        _run_test("aa-aa-aa", ("aa", "aa", "aa"))

    def test_get_account_username_split_invalid_users(self):
        """
        Ensure it fails as expected for non unlinked account users
        """

        def _run_test(username):
            try:
                result = get_account_username_split(username)
                self.fail(
                    "get_account_username_split did not raise an "
                    "exception when provided username for a non unlinked "
                    "Institution.\n'%s' returned '%s'" % (username, result)
                )
            except ValueError:
                pass

        _run_test("exusername")
        _run_test("tmp-abcde")
        _run_test("ex-abc_def")
        _run_test("")


class GetRandomUsernameTests(TestCase):
    """
    Tests the get_random_username method in utils.py
    """

    def test_get_random_username_valid(self):
        """
        Ensure all outputted usernames have the 'tmp' prefix
        """
        for i in range(100):
            username = get_random_username()
            if not username.startswith("tmp"):
                self.fail(
                    "get_random_username outputted a username that "
                    "did not start with 'tmp_': %s" % username
                )

    def test_get_random_username_unique(self):
        """
        Ensure all outputted usernames are unique
        """
        usernames = []
        for i in range(100):
            username = get_random_username()
            if username in usernames:
                self.fail(
                    "get_random_username outputted a duplicate "
                    "username: %s" % username
                )
            usernames.append(username)


class GetRedirectUrlTests(TestCase):
    """
    Tests the get_redirect_url method in utils.py
    """

    factory = RequestFactory()

    def test_get_redirect_url_query_parameter(self):
        """
        Ensure method returns URL provided as query parameter
        over everything else if present
        """
        request = self.factory.get(
            "/accounts/login/", data={"next": "/next-page/"}
        )
        result = get_redirect_url(request)
        self.assertEqual(result, "/next-page/")

    def test_get_redirect_url_referer(self):
        """
        Ensure method returns the referring page if use_referer
        is True, and no relevant query parameter was provided
        """
        request = self.factory.get(
            "/accounts/cas-login/test-inst/", HTTP_REFERER="/home/"
        )
        result = get_redirect_url(
            request, use_referer=True, default_url="/default/"
        )
        self.assertEqual(result, "/home/")

    def test_get_redirect_url_defaults(self):
        """
        Ensure method uses the provided default URL or default
        login URL setting if no other options are available
        """
        request = self.factory.get("/accounts/login/", HTTP_REFERER="/home")
        result = get_redirect_url(
            request, use_referer=False, default_url="/default/"
        )
        self.assertEqual(result, "/default/")

        request = self.factory.get("/accounts/login/", HTTP_REFERER="/home")
        result = get_redirect_url(request)
        default_login = DEFAULT_SETTING_VALUES["UNIAUTH_LOGIN_REDIRECT_URL"]
        self.assertEqual(result, default_login)


class GetServiceUrlTests(TestCase):
    """
    Tests the get_service_url method in utils.py
    """

    factory = RequestFactory()

    def test_get_service_url_correct(self):
        """
        Ensure method behaves properly when redirect_url
        is or is not provided
        """
        request = self.factory.get(
            "/accounts/login/", data={"next": "/next-page/"}
        )

        result = get_service_url(request)
        expected = "http://testserver/accounts/login/?next=%2Fnext-page%2F"
        self.assertEqual(result, expected)

        result = get_service_url(request, "/alt/target/")
        expected = "http://testserver/accounts/login/?next=%2Falt%2Ftarget%2F"
        self.assertEqual(result, expected)

    def test_get_service_url_query_parameters(self):
        """
        Ensure method preserves query parameters
        """
        request = self.factory.get("/origin/?foo=bar", secure=True)
        result = get_service_url(request, "/target/")
        expected = "https://testserver/origin/?foo=bar&next=%2Ftarget%2F"
        assert_urls_equivalent(result, expected, self.assertEqual)

        request = self.factory.get(
            "/origin/",
            secure=True,
            data={"foo": "bar", "next": "/next-page/?cat=dog"},
        )
        result = get_service_url(request)
        expected = (
            "https://testserver/origin/?foo=bar"
            "&next=%2Fnext-page%2F%3Fcat%3Ddog"
        )
        assert_urls_equivalent(result, expected, self.assertEqual)

    def test_get_service_url_ignore_ticket(self):
        """
        Ensures ticket is ignored if present as query parameter
        """
        request = self.factory.get(
            "/origin/",
            secure=True,
            data={
                "foo": "bar",
                "next": "/next-page/?cat=dog",
                "ticket": "FAKE-ticket-456",
            },
        )
        result = get_service_url(request)
        expected = (
            "https://testserver/origin/?foo=bar"
            "&next=%2Fnext-page%2F%3Fcat%3Ddog"
        )
        assert_urls_equivalent(result, expected, self.assertEqual)


class GetSettingTests(TestCase):
    """
    Tests the get_setting method in utils.py
    """

    def test_get_setting_default_settings(self):
        """
        Ensure the default values for Uniauth's settings are
        returned when not provided in the settings file
        """
        # The test settings file does not set any Uniauth
        # settings - so everything should be the default value
        self.assertEqual(get_setting("LOGIN_URL"), "/accounts/login/")
        self.assertEqual(get_setting("PASSWORD_RESET_TIMEOUT_DAYS"), 3)
        self.assertEqual(get_setting("UNIAUTH_LOGIN_DISPLAY_STANDARD"), True)
        for setting, default_value in DEFAULT_SETTING_VALUES.items():
            self.assertEqual(get_setting(setting), default_value)

    @override_settings(
        LOGIN_URL="/new/login/",
        UNIAUTH_LOGIN_DISPLAY_STANDARD=False,
        UNIAUTH_MAX_LINKED_EMAILS=-11,
    )
    def test_get_setting_provided_settings(self):
        """
        Ensure the provided values for settings are used
        when present
        """
        self.assertEqual(get_setting("LOGIN_URL"), "/new/login/")
        self.assertEqual(get_setting("UNIAUTH_LOGIN_DISPLAY_STANDARD"), False)
        self.assertEqual(get_setting("UNIAUTH_MAX_LINKED_EMAILS"), -11)
        # Should still be the default value, since it wasn't set
        self.assertEqual(get_setting("PASSWORD_RESET_TIMEOUT_DAYS"), 3)

    def test_get_setting_nonexistent_settings(self):
        """
        Ensure it fails for settings that don't exist
        """

        def _run_test(setting):
            try:
                result = get_setting(setting)
                self.fail(
                    "get_setting did not raise an exception when"
                    "asked for non-existent setting: '%s' returned '%s'"
                    % (setting, result)
                )
            except KeyError:
                pass

        _run_test("dne")
        _run_test("UNIAUTH_FAKE_SETTING")
        _run_test("")


class IsTmpUserTests(TestCase):
    """
    Tests the is_tmp_user method in utils.py
    """

    def test_is_tmp_user_profile(self):
        """
        Ensure users with verified Uniauth profiles return False
        """
        user1 = User.objects.create(username="test@example.com")
        user2 = User.objects.create(username="abc@def.ghi_0002")
        users = [user1, user2]

        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=True):
            for user in users:
                self.assertFalse(is_tmp_user(user))
        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=False):
            for user in users:
                self.assertFalse(is_tmp_user(user))

    def test_is_tmp_user_institution_account(self):
        """
        Ensure users logged in via an Insitution Account return
        True if standalone accounts aren't allowed, False otherwise
        """
        user1 = User.objects.create(username="cas-princeton-netid")
        user2 = User.objects.create(username="cas-long-dashed-slug-id123")
        users = [user1, user2]

        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=True):
            for user in users:
                self.assertFalse(is_tmp_user(user))
        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=False):
            for user in users:
                self.assertTrue(is_tmp_user(user))

    def test_is_tmp_user_temporary_user(self):
        """
        Ensure users with temporary usernames return True
        """
        user1 = User.objects.create(username="tmp-example")
        user2 = User.objects.create(username="tmp-20190217200307725912_wGgOE")
        users = [user1, user2]

        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=True):
            for user in users:
                self.assertTrue(is_tmp_user(user))
        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=False):
            for user in users:
                self.assertTrue(is_tmp_user(user))

    def test_is_tmp_user_unauthenticated(self):
        """
        Ensure unauthenticated users return False
        """
        user1 = User.objects.create(username="")
        user2 = AnonymousUser()
        users = [user1, user2]

        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=True):
            for user in users:
                self.assertFalse(is_tmp_user(user))
        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=False):
            for user in users:
                self.assertFalse(is_tmp_user(user))



from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import user_passes_test

from uniauth.utils import is_tmp_user


def login_required(
    function=None, redirect_field_name=REDIRECT_FIELD_NAME, login_url=None
):
    """
    Replacement for django's built-in login_required
    decorator that also requires user to not be a
    temporary user (must have completed signup process).

    It can be used identically to the built-in version.
    """
    actual_decorator = user_passes_test(
        lambda u: u.is_authenticated and not is_tmp_user(u),
        login_url=login_url,
        redirect_field_name=redirect_field_name,
    )
    if function:
        return actual_decorator(function)
    return actual_decorator



from django.test import TestCase, override_settings


class SimpleTestCase(TestCase):
    """
    Tests that the testing framework itself is setup appropiately
    """

    def test_simple(self):
        """
        Ensure the assert statements work as expected
        """
        self.assertTrue(True, "True is not True?!")
        self.assertFalse(False, "False is not False?!")

    def test_imports(self):
        """
        Ensure we can import uniauth + third party dependencies
        """
        import requests
        from cas import CASClient

        from uniauth.models import UserProfile

        user = UserProfile()
        self.assertTrue(user is not None)

    @override_settings(LOGIN_URL="/simple/login/")
    def test_settings(self):
        """
        Ensure the test settings module works as expected
        """
        from django.conf import settings

        self.assertTrue(settings.TESTING)
        self.assertEqual(settings.LOGIN_URL, "/simple/login/")
        with self.settings(UNIAUTH_LOGIN_DISPLAY_STANDARD=False):
            self.assertFalse(settings.UNIAUTH_LOGIN_DISPLAY_STANDARD)



from django.contrib.auth.tokens import PasswordResetTokenGenerator
from django.utils.module_loading import import_string
from rest_framework_simplejwt.settings import api_settings as jwt_settings
from rest_framework_simplejwt.tokens import RefreshToken


class EmailVerificationTokenGenerator(PasswordResetTokenGenerator):
    """
    Creates tokens for linked email verification.
    """

    def _make_hash_value(self, email, timestamp):
        return str(email.pk) + str(email.is_verified) + str(timestamp)


token_generator = EmailVerificationTokenGenerator()


def get_jwt_tokens_for_user(user, **kwargs):
    """
    Generates a refresh token for the valid user
    """
    try:
        refresh = import_string(
            jwt_settings.TOKEN_OBTAIN_SERIALIZER
        ).get_token(user)
    except (AttributeError, ImportError) as error:
        # simplejwt defines a default token serializer that uses
        # RefreshToken, but this is here as a fallback in case something
        # is weirdly configured, or the installed simplejwt package is
        # too old to support custom serializers
        refresh = RefreshToken.for_user(user)

    return str(refresh), str(refresh.access_token)



# Generated by Django 3.1.5 on 2022-11-07 23:53

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("uniauth", "0002_auto_20190707_2305"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="institutionaccount",
            unique_together={("institution", "cas_id")},
        ),
    ]



try:
    from django.conf.urls import url
except ImportError:
    from django.urls import re_path as url

from django.core.exceptions import ImproperlyConfigured

from uniauth import views
from uniauth.utils import get_setting

# The standard login form requires URLs that don't exist when
# cas_only is used. Ensure that this setting is False.
if get_setting("UNIAUTH_LOGIN_DISPLAY_STANDARD"):
    err_msg = (
        "'uniauth.urls.cas_only' can not be used when %s is True. "
        "Please include 'uniauth.urls' instead."
    ) % "UNIAUTH_LOGIN_DISPLAY_STANDARD"
    raise ImproperlyConfigured(err_msg)

app_name = "uniauth"

urlpatterns = [
    url(r"^login/$", views.login, name="login"),
    url(
        r"^cas-login/(?P<institution>[a-z0-9\-]+)/$",
        views.cas_login,
        name="cas-login",
    ),
    url(r"^logout/$", views.logout, name="logout"),
]



from unittest import skipUnless

from django.contrib.auth.models import User
from django.test import TestCase
from rest_framework_simplejwt.tokens import AccessToken, RefreshToken

from uniauth.tokens import get_jwt_tokens_for_user

try:
    import mock
except ImportError:
    from unittest import mock


MIN_DJANGO_CUSTOM_SERIALIZER_VERSION = "2.0.0"
MIN_SIMPLEJWT_CUSTOM_SERIALIZER_VERSION = "5.1.0"


def custom_serializers_supported():
    """
    Returns whether the installed simplejwt package supports custom
    serializers or not
    """
    import django
    import rest_framework_simplejwt

    # Simple JWT also requires Django 2.0+ in order to work properly,
    # but that is not currently enforced in its requirements
    return (django.__version__ >= MIN_DJANGO_CUSTOM_SERIALIZER_VERSION) and (
        rest_framework_simplejwt.__version__
        >= MIN_SIMPLEJWT_CUSTOM_SERIALIZER_VERSION
    )


if custom_serializers_supported():
    from rest_framework_simplejwt.serializers import TokenObtainPairSerializer

    class CustomTokenObtainPairSerializer(TokenObtainPairSerializer):
        """
        Example TokenObtainPairSerializer subclass that adds custom claims
        """

        @classmethod
        def get_token(cls, user):
            token = super().get_token(user)
            token["foo"] = "bar"
            return token


class GetJWTTokensForUserTests(TestCase):
    """
    Tests the get_jwt_tokens_for_user method in tokens.py
    """

    def setUp(self):
        self.user = User.objects.create(username="johndoe@school.edu")
        self.user2 = User.objects.create(username="janedoe@gmail.com")

    def test_get_jwt_tokens_for_user_success(self):
        """
        Ensure get_jwt_tokens_for_user returns strings that decode
        into valid Refresh and Access tokens
        """
        refresh_encoded, access_encoded = get_jwt_tokens_for_user(self.user)
        refresh = RefreshToken(token=refresh_encoded)
        access = AccessToken(token=access_encoded)
        self.assertEqual(refresh["user_id"], self.user.id)
        self.assertEqual(access["user_id"], self.user.id)

        refresh_encoded, access_encoded = get_jwt_tokens_for_user(self.user2)
        refresh = RefreshToken(token=refresh_encoded)
        access = AccessToken(token=access_encoded)
        self.assertEqual(refresh["user_id"], self.user2.id)
        self.assertEqual(access["user_id"], self.user2.id)

    @skipUnless(
        custom_serializers_supported(),
        (
            "django >= '{}' and simple_jwt >= '{}' required "
            "to support custom token serializers"
        ).format(
            MIN_DJANGO_CUSTOM_SERIALIZER_VERSION,
            MIN_SIMPLEJWT_CUSTOM_SERIALIZER_VERSION,
        ),
    )
    @mock.patch(
        "uniauth.tokens.jwt_settings.TOKEN_OBTAIN_SERIALIZER",
        "tests.test_tokens.CustomTokenObtainPairSerializer",
    )
    def test_get_jwt_tokens_for_user_custom_serializer(self):
        """
        If a custom token obtain serializer is defined, ensure
        get_jwt_tokens_for_user uses that custom class
        """
        refresh_encoded, access_encoded = get_jwt_tokens_for_user(self.user)
        refresh = RefreshToken(token=refresh_encoded)
        access = AccessToken(token=access_encoded)
        self.assertEqual(refresh["user_id"], self.user.id)
        self.assertEqual(access["user_id"], self.user.id)
        self.assertEqual(refresh["foo"], "bar")
        self.assertEqual(access["foo"], "bar")



from django.contrib.auth.models import User
from django.test import RequestFactory, TestCase, override_settings

from uniauth.backends import (
    CASBackend,
    LinkedEmailBackend,
    UsernameOrLinkedEmailBackend,
)
from uniauth.models import Institution, InstitutionAccount, LinkedEmail

try:
    import mock
except ImportError:
    from unittest import mock


class CASBackendTests(TestCase):
    """
    Tests the CASBackend in backends.py
    """

    def setUp(self):
        self.factory = RequestFactory()
        self.inst = Institution.objects.create(
            name="Test Inst",
            slug="test-inst",
            cas_server_url="https://fed.testinst.edu/",
        )
        self.inst2 = Institution.objects.create(
            name="Other Inst",
            slug="other-inst",
            cas_server_url="https://fed.other.edu/",
        )

    def _get_request(self):
        request = self.factory.get("/accounts/cas-login/")
        request.session = {}
        return request

    @mock.patch("cas.CASClientV2.verify_ticket")
    def test_cas_backend_verification_failed(self, mock_verify_ticket):
        """
        Ensure backend returns None if the ticket fails to verify
        """
        mock_verify_ticket.return_value = (None, {}, None)
        backend = CASBackend()
        request = self._get_request()

        prev_num_users = User.objects.count()
        user = backend.authenticate(
            request,
            institution=self.inst,
            ticket="fake-ticket",
            service="http://www.service.com/",
        )
        self.assertEqual(user, None)
        self.assertEqual(User.objects.count(), prev_num_users)

    @mock.patch("cas.CASClientV2.verify_ticket")
    def test_cas_backend_create_user(self, mock_verify_ticket):
        """
        Ensure backend properly creates a new user when necessary
        """
        mock_verify_ticket.return_value = (
            "newuser",
            {"ticket": "fake-ticket", "service": "http://www.service.com/"},
            None,
        )
        backend = CASBackend()
        request = self._get_request()

        self.assertFalse(User.objects.filter(username="newuser").exists())
        prev_num_users = User.objects.count()
        user = backend.authenticate(
            request,
            institution=self.inst,
            ticket="fake-ticket",
            service="http://www.service.com/",
        )
        self.assertNotEqual(user, None)
        self.assertEqual(User.objects.count(), prev_num_users + 1)
        self.assertTrue(
            User.objects.filter(username="cas-test-inst-newuser").exists()
        )
        self.assertEqual(
            request.session["attributes"],
            {"ticket": "fake-ticket", "service": "http://www.service.com/"},
        )

    @mock.patch("cas.CASClientV2.verify_ticket")
    def test_cas_backend_existing_user(self, mock_verify_ticket):
        """
        Ensure backend returns existing user if possible
        """
        backend = CASBackend()
        user1 = User.objects.create(
            username="johndoe@gmail.com", email="john.doe@aol.com"
        )
        InstitutionAccount.objects.create(
            profile=user1.uniauth_profile,
            institution=self.inst,
            cas_id="john123",
        )
        user2 = User.objects.create(username="cas-test-inst-jane987")
        fakeout = User.objects.create(
            username="fakeout@gmail.com", email="fakeout@verizon.net"
        )
        InstitutionAccount.objects.create(
            profile=fakeout.uniauth_profile,
            institution=self.inst2,
            cas_id="john123",
        )
        prev_num_users = User.objects.count()

        # Test case where there is a user with an InstitutionAccount
        mock_verify_ticket.return_value = (
            "john123",
            {"ticket": "fake-ticket", "service": "http://www.example.com/"},
            None,
        )
        request = self._get_request()
        user = backend.authenticate(
            request,
            institution=self.inst,
            ticket="fake-ticket",
            service="http://www.example.com/",
        )
        self.assertNotEqual(user, None)
        self.assertEqual(user.username, "johndoe@gmail.com")
        self.assertEqual(User.objects.count(), prev_num_users)
        self.assertEqual(
            request.session["attributes"],
            {"ticket": "fake-ticket", "service": "http://www.example.com/"},
        )

        # Test case where there is an existing unlinked account user
        mock_verify_ticket.return_value = (
            "jane987",
            {"ticket": "ticket0123", "service": "http://someservice.gov/"},
            None,
        )
        request = self._get_request()
        user = backend.authenticate(
            request,
            institution=self.inst,
            ticket="ticket0123",
            service="http://someservice.gov/",
        )
        self.assertNotEqual(user, None)
        self.assertEqual(user.username, "cas-test-inst-jane987")
        self.assertEqual(User.objects.count(), prev_num_users)
        self.assertEqual(
            request.session["attributes"],
            {"ticket": "ticket0123", "service": "http://someservice.gov/"},
        )

    @mock.patch("cas.CASClientV2.verify_ticket")
    def test_cas_backend_missing_request(self, mock_verify_ticket):
        """
        Ensure authentication still works without the request parameter
        """
        backend = CASBackend()

        # Test creating a new user
        mock_verify_ticket.return_value = (
            "newuser",
            {"ticket": "fake-ticket", "service": "http://www.service.com/"},
            None,
        )
        self.assertFalse(User.objects.filter(username="newuser").exists())
        prev_num_users = User.objects.count()
        user = backend.authenticate(
            None,
            institution=self.inst,
            ticket="fake-ticket",
            service="http://www.service.com/",
        )
        self.assertNotEqual(user, None)
        self.assertEqual(User.objects.count(), prev_num_users + 1)
        self.assertTrue(
            User.objects.filter(username="cas-test-inst-newuser").exists()
        )

        # Test finding an existing user
        mock_verify_ticket.return_value = (
            "jane987",
            {"ticket": "ticket0123", "service": "http://someservice.gov/"},
            None,
        )
        User.objects.create(username="cas-test-inst-jane987")
        prev_num_users = User.objects.count()
        user = backend.authenticate(
            None,
            institution=self.inst,
            ticket="ticket0123",
            service="http://someservice.gov/",
        )
        self.assertNotEqual(user, None)
        self.assertEqual(user.username, "cas-test-inst-jane987")
        self.assertEqual(User.objects.count(), prev_num_users)

        # Test failed authentication
        mock_verify_ticket.return_value = (None, {}, None)
        prev_num_users = User.objects.count()
        user = backend.authenticate(
            None,
            institution=self.inst,
            ticket="fake-ticket",
            service="http://www.service.com/",
        )
        self.assertEqual(user, None)
        self.assertEqual(User.objects.count(), prev_num_users)


class EmailBackendTests(TestCase):
    """
    Parent class for the *EmailBackendTests
    """

    def setUp(self):
        self.factory = RequestFactory()
        self.john = User.objects.create_user(
            username="johndoe@gmail.com",
            email="johndoe@gmail.com",
            password="johnpass",
        )
        self.mary = User.objects.create_user(
            username="marysue@outlook.com",
            email="mary.sue@gmail.com",
            password="marypass",
        )
        LinkedEmail.objects.create(
            profile=self.mary.uniauth_profile,
            address="alternate@gmail.com",
            is_verified=True,
        )
        LinkedEmail.objects.create(
            profile=self.mary.uniauth_profile,
            address="pending@gmail.com",
            is_verified=False,
        )
        self.cas = User.objects.create_user(username="cas-inst-netid123")
        self.tmp = User.objects.create_user(
            username="tmp-0123_456", password="tmppass"
        )
        self.perm = User.objects.create_user(
            username="permid555", password="permpass"
        )
        self.mr = User.objects.create_user(
            username="mrjones@gmail.com",
            email="mrjones@gmail.com",
            password="mrpass",
        )
        self.ms = User.objects.create_user(
            username="msjones@gmail.com",
            email="msjones@gmail.com",
            password="mspass",
        )
        LinkedEmail.objects.create(
            profile=self.mr.uniauth_profile,
            address="sharedjones@gmail.com",
            is_verified=True,
        )
        LinkedEmail.objects.create(
            profile=self.ms.uniauth_profile,
            address="sharedjones@gmail.com",
            is_verified=True,
        )


@override_settings(UNIAUTH_ALLOW_SHARED_EMAILS=True)
class LinkedEmailBackendTests(EmailBackendTests):
    """
    Tests the LinkedEmailBackend in backends.py
    """

    def test_linked_email_backend_valid_credentials(self):
        """
        Ensure backend returns appropriate user when valid
        credentials are provided
        """
        backend = LinkedEmailBackend()

        # Log in with email + password
        user = backend.authenticate(
            None, email="johndoe@gmail.com", password="johnpass"
        )
        self.assertEqual(user, self.john)

        # Log in with email (under "username" parameter) + password
        user = backend.authenticate(
            None, username="mary.sue@gmail.com", password="marypass"
        )
        self.assertEqual(user, self.mary)

        # Log in with verified linked email
        user = backend.authenticate(
            None, email="alternate@gmail.com", password="marypass"
        )
        self.assertEqual(user, self.mary)

        # Log in with shared email address
        user = backend.authenticate(
            None, email="sharedjones@gmail.com", password="mspass"
        )
        self.assertEqual(user, self.ms)
        user = backend.authenticate(
            None, email="sharedjones@gmail.com", password="mrpass"
        )
        self.assertEqual(user, self.mr)

    def test_linked_email_backend_invalid_credentials(self):
        """
        Ensure authentication fails when invalid credentials
        are provided
        """
        backend = LinkedEmailBackend()

        # Can't log in without username or email parameter
        user = backend.authenticate(None)
        self.assertEqual(user, None)
        user = backend.authenticate(self.factory.get("/accounts/login/"))
        self.assertEqual(user, None)

        # Can't log in with incorrect username / email
        user = backend.authenticate(None, username="dne", password="johnpass")
        self.assertEqual(user, None)
        user = backend.authenticate(
            None, email="dne@gmail.com", password="marypass"
        )
        self.assertEqual(user, None)

        # Can't log in with no password
        user = backend.authenticate(None, username="cas-inst-netid123")
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="johndoe@gmail.com")
        self.assertEqual(user, None)

        # Can't log in with incorrect password
        user = backend.authenticate(
            None, email="johndoe@gmail.com", password=None
        )
        self.assertEqual(user, None)
        user = backend.authenticate(
            None, email="johndoe@gmail.com", password=""
        )
        self.assertEqual(user, None)
        user = backend.authenticate(
            None, email="johndoe@gmail.com", password="marypass"
        )
        self.assertEqual(user, None)

        # Can't log in with username
        user = backend.authenticate(
            None, email="marysue@outlook.com", password="marypass"
        )
        self.assertEqual(user, None)
        user = backend.authenticate(
            None, username="permid555", password="permpass"
        )
        self.assertEqual(user, None)

        # Can't log in with unverified email
        user = backend.authenticate(
            None, email="pending@gmail.com", password="marypass"
        )
        self.assertEqual(user, None)

        # Can't log in as a temporary user
        user = backend.authenticate(
            None, username="tmp-0123_456", password="tmppass"
        )
        self.assertEqual(user, None)


@override_settings(UNIAUTH_ALLOW_SHARED_EMAILS=True)
class UsernameOrLinkedEmailBackendTests(EmailBackendTests):
    """
    Tests the UsernameOrLinkedEmailBackend in backends.py
    """

    def test_username_orlinked_email_backend_valid_credentials(self):
        """
        Ensure backend returns appropriate user when valid
        credentials are provided
        """
        backend = UsernameOrLinkedEmailBackend()

        # Log in with email + password
        user = backend.authenticate(
            None, email="johndoe@gmail.com", password="johnpass"
        )
        self.assertEqual(user, self.john)

        # Log in with email (under "username" parameter) + password
        user = backend.authenticate(
            None, username="mary.sue@gmail.com", password="marypass"
        )
        self.assertEqual(user, self.mary)

        # Log in with verified linked email
        user = backend.authenticate(
            None, email="alternate@gmail.com", password="marypass"
        )
        self.assertEqual(user, self.mary)

        # Log in with shared email address
        user = backend.authenticate(
            None, email="sharedjones@gmail.com", password="mspass"
        )
        self.assertEqual(user, self.ms)
        user = backend.authenticate(
            None, email="sharedjones@gmail.com", password="mrpass"
        )
        self.assertEqual(user, self.mr)

        # Log in with username
        user = backend.authenticate(
            None, email="marysue@outlook.com", password="marypass"
        )
        self.assertEqual(user, self.mary)
        user = backend.authenticate(
            None, username="permid555", password="permpass"
        )
        self.assertEqual(user, self.perm)

    def test_username_or_linked_email_backend_invalid_credentials(self):
        """
        Ensure authentication fails when invalid credentials
        are provided
        """
        backend = UsernameOrLinkedEmailBackend()

        # Can't log in without username or email parameter
        user = backend.authenticate(None)
        self.assertEqual(user, None)
        user = backend.authenticate(self.factory.get("/accounts/login/"))
        self.assertEqual(user, None)

        # Can't log in with incorrect username / email
        user = backend.authenticate(None, username="dne", password="johnpass")
        self.assertEqual(user, None)
        user = backend.authenticate(
            None, email="dne@gmail.com", password="marypass"
        )
        self.assertEqual(user, None)

        # Can't log in with no password
        user = backend.authenticate(None, username="cas-inst-netid123")
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="johndoe@gmail.com")
        self.assertEqual(user, None)

        # Can't log in with incorrect password
        user = backend.authenticate(
            None, email="johndoe@gmail.com", password=None
        )
        self.assertEqual(user, None)
        user = backend.authenticate(
            None, email="johndoe@gmail.com", password=""
        )
        self.assertEqual(user, None)
        user = backend.authenticate(
            None, email="johndoe@gmail.com", password="marypass"
        )
        self.assertEqual(user, None)

        # Can't log in with unverified email
        user = backend.authenticate(
            None, email="pending@gmail.com", password="marypass"
        )
        self.assertEqual(user, None)

        # Can't log in as a temporary user
        user = backend.authenticate(
            None, username="tmp-0123_456", password="tmppass"
        )
        self.assertEqual(user, None)



"""
This command is used to remove an Institution from the database.

Execution: python manage.py remove_institution <slug>
"""

from django.core.management.base import BaseCommand, CommandError

from uniauth.models import Institution
from uniauth.utils import get_input


class Command(BaseCommand):
    help = "Removes an institution from the database."

    def add_arguments(self, parser):
        parser.add_argument("slug")

    def handle(self, *args, **options):
        slug = options["slug"]

        try:
            institution = Institution.objects.get(slug=slug)
        except Institution.DoesNotExist:
            raise CommandError("No institution with slug '%s' exists." % slug)

        answer = get_input(
            "Are you sure you want to delete institution '"
            + str(institution)
            + "'?\nThis will also delete all "
            + "InstitutionAccounts for that institution.\nAnswer [y/n]: "
        )
        if answer == "y" or answer == "yes":
            institution.delete()
            self.stdout.write("Deleted institution '%s'.\n" % str(institution))
        else:
            self.stdout.write("Canceled.\n")



try:
    from django.conf.urls import url
except ImportError:
    from django.urls import re_path as url
from uniauth import views

app_name = "uniauth"

urlpatterns = [
    url(r"^login/$", views.login, name="login"),
    url(
        r"^cas-login/(?P<institution>[a-z0-9\-]+)/$",
        views.cas_login,
        name="cas-login",
    ),
    url(r"^logout/$", views.logout, name="logout"),
    url(r"^signup/$", views.signup, name="signup"),
    url(r"^settings/$", views.settings, name="settings"),
    url(r"^link-to-profile/$", views.link_to_profile, name="link-to-profile"),
    url(
        r"^link-from-profile/(?P<institution>[a-z0-9\-]+)/$",
        views.link_from_profile,
        name="link-from-profile",
    ),
    url(
        r"^verify-token/(?P<pk_base64>[0-9A-Za-z_\-]+)/(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,32})/$",
        views.verify_token,
        name="verify-token",
    ),
    url(
        r"^password-reset/",
        views.PasswordReset.as_view(),
        name="password-reset",
    ),
    url(
        r"^password-reset-done/",
        views.PasswordResetDone.as_view(),
        name="password-reset-done",
    ),
    url(
        r"^password-reset-verify/(?P<uidb64>[0-9A-Za-z_\-]+)/(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,32})/",
        views.PasswordResetVerify.as_view(),
        name="password-reset-verify",
    ),
    url(
        r"^password-reset-verify-done/",
        views.PasswordResetVerifyDone.as_view(),
        name="password-reset-verify-done",
    ),
    url(r"^jwt-tokens/", views.get_jwt_tokens_from_session, name="jwt-tokens"),
]



from cas import CASClient
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.db.models import Q

from uniauth.models import InstitutionAccount, UserProfile
from uniauth.utils import is_tmp_user


class CASBackend(ModelBackend):
    """
    Authentication backend that verifies A CAS ticket
    with the server for the provided institution and
    returns either the authenticated Uniauth user (if
    one already exists), or a newly created user with
    a temporary username otherwise.
    """

    def authenticate(self, request, institution, ticket, service):
        user_model = get_user_model()

        # Attempt to verify the ticket with the institution's CAS server
        client = CASClient(
            version=2,
            service_url=service,
            server_url=institution.cas_server_url,
        )
        username, attributes, pgtiou = client.verify_ticket(ticket)

        # Add the attributes returned by the CAS server to the session
        if request and attributes:
            request.session["attributes"] = attributes

        # If no username was returned, verification failed
        if not username:
            return None

        # Attempt to find a user possessing an account
        # with that username for the institution
        try:
            user = InstitutionAccount.objects.get(
                cas_id=username, institution=institution
            ).profile.user
        except InstitutionAccount.DoesNotExist:
            user = None

        # If such a user does not exist, get or create
        # one with a deterministic, CAS username
        if not user:
            temp_username = "cas-%s-%s" % (institution.slug, username)
            user, created = user_model._default_manager.get_or_create(
                **{user_model.USERNAME_FIELD: temp_username}
            )

        return user


class LinkedEmailBackend(ModelBackend):
    """
    Authentication backend allowing users to authenticate
    with any email address linked to the account, along
    with their password.

    Note: 'username' is still supported as an argument for
    authenticate to facilitate easier swapping of this
    backend with the UsernameOrLinkedEmailBackend, and
    other authentication schemes using a username argument.
    """

    def _get_users(self, user_model, email):
        """
        Query for users with a verified linked email
        address matching the provided email value
        """
        return user_model._default_manager.filter(
            uniauth_profile__linked_emails__address__iexact=email,
            uniauth_profile__linked_emails__is_verified=True,
        ).all()

    def authenticate(self, request, email=None, password=None, **kwargs):
        user_model = get_user_model()

        # If email field was not provided, check for
        # alternative names, or for a "username" field
        if email is None:
            email = kwargs.get("email_address")
            if email is None:
                email = kwargs.get("username")

                # If a custom user model is being used, check for
                # the email in the specified username field
                if email is None:
                    email = kwargs.get(user_model.USERNAME_FIELD)

        # Get the user(s) who own the provided email address
        users = self._get_users(user_model, email)

        # If there were no matching users, run the password
        # hasher once, to guard against timing attacks
        if not users:
            user_model().set_password(password)
            return None

        # Otherwise, check the password for each matched user
        for user in users:
            if user.check_password(password):
                return user
        return None


class UsernameOrLinkedEmailBackend(LinkedEmailBackend):
    """
    Authenticaton backend allowing users to authenticate
    with their username, or any email address linked to
    the account, along with their password.
    """

    def _get_users(self, user_model, username):
        """
        Query for verified users with a username or linked
        email address matching the provided username value
        """
        username_field = user_model.USERNAME_FIELD
        matched_users = user_model._default_manager.filter(
            (Q(**{username_field: username}))
            | (
                Q(uniauth_profile__linked_emails__address__iexact=username)
                & Q(uniauth_profile__linked_emails__is_verified=True)
            )
        ).all()
        return filter(lambda x: not is_tmp_user(x), matched_users)



"""
Django settings for testing.
"""

import os

# Uniauth requires an actual email configuration to be set
# up (to send emails for email validation, changing passwords,
# etc.). This backend just stores emails in django.core.mail.outbox
EMAIL_BACKEND = "django.core.mail.backends.locmem.EmailBackend"

# Definitely change these for real applications
TESTING = True
SECRET_KEY = "FAKE_SECRET"
ALLOWED_HOSTS = ["*"]

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "uniauth",
    "tests",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

# Use the Uniauth authentication backends
AUTHENTICATION_BACKENDS = [
    "uniauth.backends.LinkedEmailBackend",
    "uniauth.backends.CASBackend",
]

ROOT_URLCONF = "tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "demo_app", "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

# Database
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

# Password validation
AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

# Internationalization
LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_L10N = True
USE_TZ = True

# Static files
STATIC_URL = "/static/"



from cas import CASClient
from django.contrib.auth import (
    REDIRECT_FIELD_NAME,
    authenticate,
    get_user_model,
)
from django.contrib.auth import login as auth_login
from django.contrib.auth import logout as auth_logout
from django.contrib.auth import update_session_auth_hash
from django.contrib.auth.views import (
    PasswordResetCompleteView,
    PasswordResetConfirmView,
    PasswordResetDoneView,
    PasswordResetView,
)
from django.contrib.sites.shortcuts import get_current_site
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.core.mail import EmailMessage
from django.http import (
    Http404,
    HttpResponseBadRequest,
    HttpResponseRedirect,
    JsonResponse,
)
from django.shortcuts import render
from django.template.loader import render_to_string
from django.urls import reverse, reverse_lazy
from django.urls.exceptions import NoReverseMatch
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters
from rest_framework import status

from uniauth.decorators import login_required
from uniauth.forms import (
    AddLinkedEmailForm,
    ChangePrimaryEmailForm,
    LinkedEmailActionForm,
    LoginForm,
    PasswordChangeForm,
    PasswordResetForm,
    SetPasswordForm,
    SignupForm,
)
from uniauth.merge import merge_model_instances
from uniauth.models import Institution, InstitutionAccount, LinkedEmail
from uniauth.tokens import get_jwt_tokens_for_user, token_generator
from uniauth.utils import (
    choose_username,
    decode_pk,
    encode_pk,
    get_account_username_split,
    get_protocol,
    get_random_username,
    get_redirect_url,
    get_service_url,
    get_setting,
    is_tmp_user,
    is_unlinked_account,
)

try:
    from urllib import urlencode

    from urlparse import urlunparse
except ImportError:
    from urllib.parse import urlencode, urlunparse


def _get_global_context(request):
    """
    Returns a dictionary of context data shared by the views
    """
    context = {}

    # Returns the reverse lookup URL of the provided view with
    # the provided slug argument. Returns None if the view is
    # not accessible under the current configuration
    def get_reversed_url_or_none(view_name, slug):
        try:
            return reverse("uniauth:%s" % view_name, args=[slug])
        except NoReverseMatch:
            return None

    # Add a list of Institution tuples, with each containing:
    # (name, slug, CAS login url, Profile link URL)
    institutions = Institution.objects.all()
    institutions = list(
        map(
            lambda x: (
                x.name,
                x.slug,
                get_reversed_url_or_none("cas-login", x.slug),
                get_reversed_url_or_none("link-from-profile", x.slug),
            ),
            institutions,
        )
    )
    context["institutions"] = institutions

    # Add the query parameters, as a string
    query_params = urlencode(request.GET)
    prefix = "?" if query_params else ""
    context["query_params"] = prefix + query_params

    return context


def _login_success(request, user, next_url, drop_params=[]):
    """
    Determines where to go upon successful authentication:
    Redirects to link tmp account page if user is a temporary
    user, redirects to next_url otherwise.

    Any query parameters whose key exists in drop_params are
    not propogated to the destination URL
    """
    query_params = request.GET.copy()
    jwt_auth = get_setting("UNIAUTH_USE_JWT_AUTH")

    # Drop all blacklisted query parameters
    for key in drop_params:
        if key in query_params:
            del query_params[key]

    # Temporary users should redirect to the link-to-profile page
    if is_tmp_user(user):
        query_params[REDIRECT_FIELD_NAME] = next_url
        return HttpResponseRedirect(
            reverse("uniauth:link-to-profile") + "?" + urlencode(query_params)
        )

    # All other users should redirect to next_url
    else:
        suffix = ""
        if REDIRECT_FIELD_NAME in query_params:
            del query_params[REDIRECT_FIELD_NAME]
        if len(query_params) > 0:
            suffix = "?" + urlencode(query_params)
        if jwt_auth:
            refresh, access = get_jwt_tokens_for_user(user)
            request.session["jwt-refresh"] = refresh
            request.session["jwt-access"] = access
        return HttpResponseRedirect(next_url + suffix)


def login(request):
    """
    Authenticates the user, then redirects them to the
    next page, defaulting to the URL specified by the
    UNIAUTH_LOGIN_REDIRECT_URL setting.

    Offers users the choice between logging in with their
    Uniauth credentials, or through the CAS interface for
    any supported institution.
    """
    next_url = request.GET.get("next")
    context = _get_global_context(request)

    if not next_url:
        next_url = get_redirect_url(request)

    # If the user is already authenticated, proceed to next page
    if request.user.is_authenticated:
        return _login_success(request, request.user, next_url)

    display_standard = get_setting("UNIAUTH_LOGIN_DISPLAY_STANDARD")
    display_cas = get_setting("UNIAUTH_LOGIN_DISPLAY_CAS")
    num_institutions = len(context["institutions"])

    # Ensure the login settings are configured correctly
    if not display_standard and not display_cas:
        err_msg = "At least one of '%s' and '%s' must be True." % (
            "UNIAUTH_LOGIN_DISPLAY_STANDARD",
            "UNIAUTH_LOGIN_DISPLAY_CAS",
        )
        raise ImproperlyConfigured(err_msg)
    if display_cas and num_institutions == 0:
        err_msg = (
            "'%s' is True, but there are no Institutions in the "
            "database to log into!"
        ) % "UNIAUTH_LOGIN_DISPLAY_CAS"
        raise ImproperlyConfigured(err_msg)

    context["display_standard"] = display_standard
    context["display_cas"] = display_cas
    context["num_institutions"] = num_institutions

    # If we aren't displaying the standard login form,
    # we're just displaying the CAS login options
    if not display_standard:
        institutions = context["institutions"]
        query_params = context["query_params"]

        # If there's only one possible institution to log
        # into, redirect to its CAS login page immediately
        if num_institutions == 1:
            return HttpResponseRedirect(institutions[0][2] + query_params)

        # Otherwise, render the page (without the Login form)
        else:
            return render(request, "uniauth/login.html", context)

    # If we are displaying the login form, and it's
    # a POST request, attempt to validate the form
    elif request.method == "POST":
        form = LoginForm(request, request.POST)

        # Authentication successful: setup session + proceed
        if form.is_valid():
            user = form.get_user()
            auth_login(request, user)
            request.session["auth-method"] = "uniauth"
            return _login_success(request, user, next_url)

        # Authentication failed: render form errors
        else:
            context["form"] = form
            return render(request, "uniauth/login.html", context)

    # Otherwise, render a blank Login form
    else:
        form = LoginForm(request)
        context["form"] = form
        return render(request, "uniauth/login.html", context)


def cas_login(request, institution):
    """
    Redirects to the CAS login URL, or verifies the
    CAS ticket, if provided.

    Accepts the slug of the institution to log in to.
    """
    next_url = request.GET.get("next")
    ticket = request.GET.get("ticket")

    # Ensure there is an institution with the provided slug
    try:
        institution = Institution.objects.get(slug=institution)
    except Institution.DoesNotExist:
        raise Http404

    if not next_url:
        next_url = get_redirect_url(request, use_referer=True)

    # If the user is already authenticated, proceed to next page
    if request.user.is_authenticated:
        return _login_success(request, request.user, next_url)

    service_url = get_service_url(request, next_url)
    client = CASClient(
        version=2,
        service_url=service_url,
        server_url=institution.cas_server_url,
    )

    # If a ticket was provided, attempt to authenticate with it
    if ticket:
        user = authenticate(
            request=request,
            institution=institution,
            ticket=ticket,
            service=service_url,
        )

        # Authentication successful: setup session + proceed
        if user:
            if not request.session.exists(request.session.session_key):
                request.session.create()
            auth_login(request, user)
            request.session["auth-method"] = "cas-" + institution.slug
            return _login_success(request, user, next_url, ["ticket"])

        # Authentication failed: raise permission denied
        else:
            raise PermissionDenied("Verification of CAS ticket failed.")

    # If no ticket was provided, redirect to the
    # login URL for the institution's CAS server
    else:
        return HttpResponseRedirect(client.get_login_url())


def logout(request):
    """
    Logs the user out of their Uniauth account, and
    redirects to the next page, defaulting to the URL
    specified by the UNIAUTH_LOGOUT_REDIRECT_URL setting.

    If no redirect page is set (URL parameter not given
    and UNIAUTH_LOGOUT_REDIRECT_URL is None), renders the
    logout template.

    Also logs the user out of CAS if they logged in
    via CAS, and the UNIAUTH_LOGOUT_CAS_COMPLETELY
    setting is true.
    """
    next_page = request.GET.get("next")
    auth_method = request.session.get("auth-method")

    if not next_page and get_setting("UNIAUTH_LOGOUT_REDIRECT_URL"):
        next_page = get_redirect_url(
            request, get_setting("UNIAUTH_LOGOUT_REDIRECT_URL")
        )

    # Formally log out user
    auth_logout(request)

    # Determine whether the user logged in through an institution's CAS
    institution = None
    if auth_method and auth_method.startswith("cas-"):
        try:
            institution = Institution.objects.get(slug=auth_method[4:])
        except Institution.DoesNotExist:
            pass

    # If we need to logout an institution's CAS,
    # redirect to that CAS server's logout URL
    if institution and get_setting("UNIAUTH_LOGOUT_CAS_COMPLETELY"):
        redirect_url = urlunparse(
            (
                get_protocol(request),
                request.get_host(),
                next_page or reverse("uniauth:logout"),
                "",
                "",
                "",
            )
        )
        client = CASClient(
            version=2,
            service_url=get_service_url(request),
            server_url=institution.cas_server_url,
        )
        return HttpResponseRedirect(client.get_logout_url(redirect_url))

    # If next page is set, proceed to it
    elif next_page:
        return HttpResponseRedirect(next_page)

    # Otherwise, render the logout view
    else:
        return render(request, "uniauth/logout.html")


def _send_verification_email(request, to_email, verify_email):
    """
    Sends an email (to to_email) containing a link to
    verify the target email (verify_email) as a linked
    email address for the user.

    Expects to_email as a string, and verify_email as
    a LinkedEmail model instance.
    """
    subject = "Verify your email address."
    message = render_to_string(
        "uniauth/verification-email.html",
        {
            "protocol": get_protocol(request),
            "domain": get_current_site(request),
            "pk": encode_pk(verify_email.pk),
            "token": token_generator.make_token(verify_email),
            "query_params": _get_global_context(request)["query_params"],
        },
    )
    email = EmailMessage(
        subject,
        message,
        to=[to_email],
        from_email=get_setting("UNIAUTH_FROM_EMAIL"),
    )
    email.send()


def signup(request):
    """
    Creates a new Uniauth profile with the provided
    primary email address and password.

    Prompts user to verify the email address before
    profile is fully created.
    """
    next_url = request.GET.get("next")
    context = _get_global_context(request)

    if not next_url:
        next_url = get_redirect_url(request)

    # If the user is already authenticated + has a Uniauth
    # profile, proceed to next page
    if (
        request.user.is_authenticated
        and not is_tmp_user(request.user)
        and not is_unlinked_account(request.user)
    ):
        return HttpResponseRedirect(next_url)

    # If it's a POST request, attempt to validate the form
    if request.method == "POST":
        form = SignupForm(request.POST)

        # Validation successful: setup temporary user
        if form.is_valid():
            form_email = form.cleaned_data["email"]
            user = request.user

            # If the user is not already authenticated, create a User
            if not user or not user.is_authenticated:
                tmp_username = get_random_username()
                user_model = get_user_model()
                user = user_model.objects.create(username=tmp_username)

            # Set user's password + create linked email
            user.set_password(form.cleaned_data["password1"])
            user.save()
            email, _ = LinkedEmail.objects.get_or_create(
                profile=user.uniauth_profile,
                address=form_email,
                is_verified=False,
            )

            # Send verification email + render waiting template
            _send_verification_email(request, email.address, email)
            return render(
                request,
                "uniauth/verification-waiting.html",
                {"email": email.address, "is_signup": True},
            )

        # Validation failed: render form errors
        else:
            context["form"] = form
            return render(request, "uniauth/signup.html", context)

    # Otherwise, render a blank Signup form
    else:
        form = SignupForm()
        context["form"] = form
        return render(request, "uniauth/signup.html", context)


@login_required
def settings(request):
    """
    Allows the user to link additional emails, change
    the primary email address of, and link additional
    institution accounts to their Uniauth profile.
    """
    context = _get_global_context(request)
    context["email_resent"] = None
    context["email_added"] = None
    context["password_changed"] = False

    action_email_form = None
    add_email_form = None
    change_email_form = None
    change_password_form = None

    # This page may only be accessed by users with Uniauth profiles:
    # if the user is logged in with an unlinked InsitutionAccount,
    # redirect them to the link page
    if is_unlinked_account(request.user):
        params = urlencode({"next": reverse("uniauth:settings")})
        return HttpResponseRedirect(
            reverse("uniauth:link-to-profile") + "?" + params
        )

    # If it's a POST request, determine which form was submitted
    if request.method == "POST":

        # Linked Email Action Form submitted: validate the
        # delete / resend action + perform it if able
        if request.POST.get("action-email-submitted"):
            action_email_form = LinkedEmailActionForm(request.POST)
            if action_email_form.is_valid():
                delete_pk = action_email_form.cleaned_data.get("delete_pk")
                resend_pk = action_email_form.cleaned_data.get("resend_pk")

                # If they asked to delete a linked email: it must
                # belong to the user and not be the primary address
                if delete_pk:
                    try:
                        email = LinkedEmail.objects.get(pk=delete_pk)
                        if (
                            email.profile.user == request.user
                            and email.address != request.user.email
                        ):
                            email.delete()
                    except LinkedEmail.DoesNotExist:
                        pass

                # If they asked to resend a verification email: it
                # must belong to the user and be pending verification
                elif resend_pk:
                    try:
                        email = LinkedEmail.objects.get(pk=resend_pk)
                        if (
                            email.profile.user == request.user
                            and not email.is_verified
                        ):
                            _send_verification_email(
                                request, email.address, email
                            )
                            context["email_resent"] = email.address
                    except LinkedEmail.DoesNotExist:
                        pass

                action_email_form = None

        # Add Linked Email Form submitted: validate and create
        # new linked email + send verification email
        elif request.POST.get("add-email-submitted"):
            add_email_form = AddLinkedEmailForm(request.user, request.POST)
            if add_email_form.is_valid():
                email = LinkedEmail.objects.create(
                    profile=request.user.uniauth_profile,
                    address=add_email_form.cleaned_data["email"],
                    is_verified=False,
                )
                _send_verification_email(request, email.address, email)
                context["email_added"] = email.address
                add_email_form = None

        # Change Primary Email Address Form submitted:
        # validate and set user's primary email address
        elif request.POST.get("change-email-submitted"):
            change_email_form = ChangePrimaryEmailForm(
                request.user, request.POST
            )
            if change_email_form.is_valid():
                request.user.email = change_email_form.cleaned_data["email"]
                request.user.save()
                change_email_form = None

        # Change Password Form submitted: validate and set password
        elif request.POST.get("change-password-submitted"):
            change_password_form = PasswordChangeForm(
                request.user, request.POST
            )
            if change_password_form.is_valid():
                change_password_form.save()
                update_session_auth_hash(request, request.user)
                context["password_changed"] = True
                change_password_form = None

    # Create blank versions of all non-bound forms
    if action_email_form is None:
        action_email_form = LinkedEmailActionForm(request.user)
    if add_email_form is None:
        add_email_form = AddLinkedEmailForm(request.user)
    if change_email_form is None:
        change_email_form = ChangePrimaryEmailForm(request.user)
    if change_password_form is None:
        change_password_form = PasswordChangeForm(request.user)

    # Assemble the context and render
    context["action_email_form"] = action_email_form
    context["add_email_form"] = add_email_form
    context["change_email_form"] = change_email_form
    context["change_password_form"] = change_password_form
    return render(request, "uniauth/settings.html", context)


def _add_institution_account(profile, slug, cas_id):
    """
    Accepts an institution slug and cas ID and links an
    InsitutionAccount to the provided Uniauth user.
    """
    institution = Institution.objects.get(slug=slug)
    InstitutionAccount.objects.create(
        profile=profile, institution=institution, cas_id=cas_id
    )


def link_to_profile(request):
    """
    If the user is a temporary one who was logged in via
    an institution (not through a Uniauth profile), offers
    them the choice between logging to an existing Uniauth
    account or creating a new one.

    The institution account is (eventually) linked to the
    Uniauth profile the user logged into / created.
    """
    next_url = request.GET.get("next")
    context = _get_global_context(request)

    if not next_url:
        next_url = get_redirect_url(request)

    params = urlencode({"next": next_url})
    context["next_url"] = next_url

    # If the user is not authenticated at all, redirect to login page
    if not request.user.is_authenticated:
        return HttpResponseRedirect(reverse("uniauth:login") + "?" + params)

    # If the user is already authenticated + verified, proceed to next page
    if not is_tmp_user(request.user) and not is_unlinked_account(request.user):
        return HttpResponseRedirect(next_url)

    # If the user is temporary, but was not logged in via an institution
    # (e.g. created through Uniauth, but not verified), redirect to signup
    if not is_unlinked_account(request.user):
        return HttpResponseRedirect(reverse("uniauth:signup") + "?" + params)

    # At this point, we've ensured the user is temporary and was
    # logged in via an institution. We just need to handle the
    # Login Form, if the user chooses to link to an existing account.

    # If it's a POST request, attempt to validate the form
    if request.method == "POST":
        form = LoginForm(request, request.POST)

        # Authentication successful
        if form.is_valid():
            unlinked_user = request.user
            username_split = get_account_username_split(request.user.username)

            # Log in as the authenticated Uniauth user
            user = form.get_user()
            auth_login(request, user)

            # Merge the unlinked account into the logged in profile,
            # then add the institution account described by the username
            merge_model_instances(user, [unlinked_user])
            _add_institution_account(
                user.uniauth_profile, username_split[1], username_split[2]
            )

            slug = username_split[1]
            context["institution"] = Institution.objects.get(slug=slug)
            return render(request, "uniauth/link-success.html", context)

        # Authentication failed: render form errors
        else:
            context["form"] = form
            return render(request, "uniauth/link-to-profile.html", context)

    # Otherwise, render a blank Login form
    else:
        form = LoginForm(request)
        context["form"] = form
        return render(request, "uniauth/link-to-profile.html", context)


def link_from_profile(request, institution):
    """
    Attempts to authenticate a CAS account for the provided
    institution, and links it to the current Uniauth profile
    if successful.
    """
    next_url = request.GET.get("next")
    ticket = request.GET.get("ticket")

    # Ensure there is an institution with the provided slug
    try:
        institution = Institution.objects.get(slug=institution)
    except Institution.DoesNotExist:
        raise Http404

    if not next_url:
        next_url = get_redirect_url(request, use_referer=True)

    # If the user is not already logged into a verified
    # Uniauth account, raise permission denied
    if (
        not request.user.is_authenticated
        or is_tmp_user(request.user)
        or is_unlinked_account(request.user)
    ):
        raise PermissionDenied("Must be logged in as verified Uniauth user.")

    service_url = get_service_url(request, next_url)
    client = CASClient(
        version=2,
        service_url=service_url,
        server_url=institution.cas_server_url,
    )

    # If a ticket was provided, attempt to authenticate with it
    if ticket:
        user = authenticate(
            request=request,
            institution=institution,
            ticket=ticket,
            service=service_url,
        )

        # Authentication successful: link to Uniauth profile if
        # the institution account has not been linked yet + proceed
        if user:
            if is_unlinked_account(user):
                merge_model_instances(request.user, [user])
                username_split = get_account_username_split(user.username)
                _add_institution_account(
                    request.user.uniauth_profile,
                    username_split[1],
                    username_split[2],
                )

            return HttpResponseRedirect(next_url)

        # Authentication failed: raise permission denied
        else:
            raise PermissionDenied("Verification of CAS ticket failed")

    # If no ticket was provided, redirect to the
    # login URL for the institution's CAS server
    else:
        return HttpResponseRedirect(client.get_login_url())


def verify_token(request, pk_base64, token):
    """
    Verifies a token generated for validating an email
    address, and notifies the user whether verification
    was successful.
    """
    next_url = request.GET.get("next") or request.GET.get(REDIRECT_FIELD_NAME)
    context = {"next_url": next_url, "is_signup": False}
    user_model = get_user_model()

    # Attempt to get the linked email to verify
    try:
        email_pk = decode_pk(pk_base64)
        email = LinkedEmail.objects.get(pk=email_pk)
    except (
        TypeError,
        ValueError,
        OverflowError,
        user_model.DoesNotExist,
        LinkedEmail.DoesNotExist,
    ):
        email = None

    # In the unlikely scenario that a user is trying to sign up
    # with an email another verified user has as a primary email
    # address, reject verification immediately
    if (
        email is not None
        and is_tmp_user(email.profile.user)
        and get_user_model().objects.filter(email=email.address).exists()
    ):
        email = None

    # If the token successfully verifies, update the linked email
    if email is not None and token_generator.check_token(email, token):
        email.is_verified = True
        email.save()

        # If the user this email is linked to is a temporary
        # one, change it to a fully registered user
        user = email.profile.user
        if is_tmp_user(user) or is_unlinked_account(user):
            context["is_signup"] = True
            old_username = user.username

            # Change the email + username to the verified email
            user.email = email.address
            user.username = choose_username(user.email)
            user.save()

            # If the user was created via CAS, add the institution
            # account described by the temporary username
            if old_username.startswith("cas"):
                username_split = get_account_username_split(old_username)
                _add_institution_account(
                    user.uniauth_profile, username_split[1], username_split[2]
                )

        # If UNIAUTH_ALLOW_SHARED_EMAILS is False, and there were
        # pending LinkedEmails for this address on other accounts,
        # delete them
        if not get_setting("UNIAUTH_ALLOW_SHARED_EMAILS"):
            LinkedEmail.objects.filter(
                address=email.address, is_verified=False
            ).delete()

        return render(request, "uniauth/verification-success.html", context)

    # If anything went wrong, just render the failed verification template
    else:
        return render(request, "uniauth/verification-failure.html", context)


# The password reset views are pulled from the django.conrib.auth
# package, and are used largely unmodified. We just set things like
# the template and reverse URL names to use Uniauth's files + naming
# scheme.


class PasswordReset(PasswordResetView):
    """
    Asks user for an email address to send the link to, then
    sends a password reset link to that address.

    If no user has that email address linked to their account,
    no email is sent, unbeknownst to the user.
    """

    email_template_name = "uniauth/password-reset-email.html"
    form_class = PasswordResetForm
    from_email = get_setting("UNIAUTH_FROM_EMAIL")
    success_url = reverse_lazy("uniauth:password-reset-done")
    template_name = "uniauth/password-reset.html"

    def form_valid(self, form):
        """
        Add global context data before composing reset email
        """
        self.extra_email_context = _get_global_context(self.request)
        return super(PasswordReset, self).form_valid(form)


class PasswordResetDone(PasswordResetDoneView):
    """
    Shows success message once email was sent.
    """

    template_name = "uniauth/password-reset-done.html"


class PasswordResetVerify(PasswordResetConfirmView):
    """
    Verifies the provided token, and prompts user for their
    new password if successful.
    """

    form_class = SetPasswordForm
    success_url = reverse_lazy("uniauth:password-reset-verify-done")
    template_name = "uniauth/password-reset-verify.html"

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        """
        Save query params in session if available.
        """
        query_params = _get_global_context(self.request)["query_params"]
        if query_params:
            self.request.session["password-reset-query-params"] = query_params
        return super(PasswordResetVerify, self).dispatch(*args, **kwargs)

    def form_valid(self, form):
        """
        Set success_url to include query params saved in session
        """
        query_params = self.request.session.pop(
            "password-reset-query-params", ""
        )
        self.success_url += query_params
        return super(PasswordResetVerify, self).form_valid(form)


class PasswordResetVerifyDone(PasswordResetCompleteView):
    """
    Shows success message once new password is set.
    """

    template_name = "uniauth/password-reset-verify-done.html"

    def get_context_data(self, **kwargs):
        """
        Add global context data to template context.
        """
        context = super(PasswordResetVerifyDone, self).get_context_data(
            **kwargs
        )
        context.update(_get_global_context(self.request))
        return context


def get_jwt_tokens_from_session(request):
    if request.method == "GET":
        refresh = request.session.pop("jwt-refresh", None)
        access = request.session.pop("jwt-access", None)
        if refresh is None or access is None:
            return JsonResponse({}, status=status.HTTP_404_NOT_FOUND)
        else:
            return JsonResponse(
                {"refresh": refresh, "access": access},
                status=status.HTTP_200_OK,
            )



from datetime import timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone


class UserProfile(models.Model):
    """
    Extension of the default User model containing
    extra information necessary for UniAuth to run.
    """

    # User this profile is extending
    user = models.OneToOneField(
        get_user_model(),
        related_name="uniauth_profile",
        on_delete=models.CASCADE,
        null=False,
    )

    def get_display_id(self):
        """
        Returns a display-friendly ID for this User, using their
        username. Users created via CAS authentication will have
        their CAS ID returned (without the cas-institution prefix),
        and Users with an email address for a username will have
        the string preceeding the "@" returned. All other users
        will have their raw username returned.

        Note that these IDs are not guaranteed to be unique.
        """
        username = self.user.username
        if "@" in username:
            return username.split("@")[0]
        if username.startswith("cas-"):
            from uniauth.utils import get_account_username_split

            return get_account_username_split(username)[-1]
        return username

    def __str__(self):
        try:
            return self.user.email or self.user.username
        except:
            return "NULL"


@receiver(post_save, sender=get_user_model())
def create_user_profile(sender, instance, created, **kwargs):
    """
    Create a Uniauth profile automatically when a User is created.

    If the User was given an email on creation, add it as a verified
    LinkedEmail immediately.
    """
    if created:
        profile = UserProfile.objects.create(user=instance)
        if profile and instance.email:
            LinkedEmail.objects.create(
                profile=profile, address=instance.email, is_verified=True
            )


@receiver(post_save, sender=get_user_model())
def clear_old_tmp_users(sender, instance, created, **kwargs):
    """
    Deletes temporary users more than PASSWORD_RESET_TIMEOUT_DAYS
    old when a User is created.

    Does nothing if the user model does not have date_joined field.
    """
    if created:
        user_model = get_user_model()
        if hasattr(user_model, "date_joined"):
            from uniauth.utils import get_setting

            timeout_days = timedelta(
                days=get_setting("PASSWORD_RESET_TIMEOUT_DAYS")
            )
            tmp_expire_date = (timezone.now() - timeout_days).replace(
                hour=0, minute=0, second=0, microsecond=0
            )
            user_model.objects.filter(
                username__startswith="tmp-", date_joined__lt=tmp_expire_date
            ).delete()


class LinkedEmail(models.Model):
    """
    Represents an email address linked to a user's account.
    """

    # Person owning this email
    profile = models.ForeignKey(
        "UserProfile",
        related_name="linked_emails",
        on_delete=models.CASCADE,
        null=False,
    )

    # The email address
    address = models.EmailField(null=False, blank=False)

    # Whether the linked email is verified
    is_verified = models.BooleanField(default=False)

    def clean(self):
        """
        Ensures an email can't be linked and verified for multiple
        accounts if UNIAUTH_ALLOW_SHARED_EMAILS is False.

        Also ensures a user does not link more than the maximum
        number of linked emails per user.
        """
        from uniauth.utils import get_setting

        # Check for shared emails if necessary
        if (
            not get_setting("UNIAUTH_ALLOW_SHARED_EMAILS")
            and LinkedEmail.objects.filter(
                address=self.address, is_verified=True
            )
            and self.is_verified
        ):
            raise ValidationError(
                "This email address has already been "
                + "linked to another account."
            )

        # Ensure a user doesn't link more than the maximum
        max_linked_emails = get_setting("UNIAUTH_MAX_LINKED_EMAILS")
        if (
            max_linked_emails > 0
            and self.profile.linked_emails.count() >= max_linked_emails
        ):
            raise ValidationError(
                ("You can not link more than %d emails " "to your account.")
                % max_linked_emails
            )
        super(LinkedEmail, self).clean()

    def __str__(self):
        try:
            return "%s | %s" % (self.profile, self.address)
        except:
            return "NULL"


class Institution(models.Model):
    """
    Represents an organization holding a CAS server
    that can be logged into.
    """

    # Name of the institution
    name = models.CharField(max_length=30, null=False, blank=False)

    # Slugified version of name
    slug = models.CharField(
        max_length=30, null=False, blank=False, unique=True
    )

    # CAS server location
    cas_server_url = models.URLField(null=False, blank=False)

    def __str__(self):
        try:
            return self.slug
        except:
            return "NULL"


class InstitutionAccount(models.Model):
    """
    Relates users to the accounts they have at
    institutions, and stores any associated data.
    """

    # The person the account is for
    profile = models.ForeignKey(
        "UserProfile",
        related_name="accounts",
        on_delete=models.CASCADE,
        null=False,
    )

    # The institution the account is with
    institution = models.ForeignKey(
        "Institution",
        related_name="accounts",
        on_delete=models.CASCADE,
        null=False,
    )

    # The ID used by the CAS server
    cas_id = models.CharField(max_length=30, null=False, blank=False)

    class Meta:
        unique_together = ("institution", "cas_id")

    def __str__(self):
        try:
            return "%s | %s | account" % (self.profile, self.institution)
        except:
            return "NULL"



from django.contrib.auth.models import User
from django.test import TestCase, override_settings

from uniauth.merge import merge_model_instances
from uniauth.models import Institution, InstitutionAccount, LinkedEmail


class MergeModelInstancesTests(TestCase):
    """
    Tests the merge_model_instances method in merge.py
    """

    def setUp(self):
        self.inst1 = Institution.objects.create(
            name="Test Uni",
            slug="test-uni",
            cas_server_url="https://cas.testuni.edu",
        )
        self.inst2 = Institution.objects.create(
            name="Other Inst",
            slug="other-inst",
            cas_server_url="https://fed.other-inst.edu",
        )

        self.user1 = User.objects.create(
            username="user1@example.com", email="user1@example.com"
        )
        self.user2 = User.objects.create(username="cas-test-uni-user2")
        InstitutionAccount.objects.create(
            profile=self.user2.uniauth_profile,
            institution=self.inst1,
            cas_id="user2",
        )

        self.user3 = User.objects.create(
            username="user3@gmail.com", email="user3@gmail.com"
        )
        LinkedEmail.objects.create(
            profile=self.user3.uniauth_profile, address="backup@gmail.com"
        )
        InstitutionAccount.objects.create(
            profile=self.user3.uniauth_profile,
            institution=self.inst1,
            cas_id="john123",
        )
        self.user4 = User.objects.create(
            username="us.er4@nowhere.gov", email="huh@what.edu"
        )
        LinkedEmail.objects.create(
            profile=self.user4.uniauth_profile,
            address="another@gmail.com",
            is_verified=True,
        )
        InstitutionAccount.objects.create(
            profile=self.user4.uniauth_profile,
            institution=self.inst2,
            cas_id="exid001",
        )
        self.user5 = User.objects.create(
            username="user5@foo.bar", email="user5@foo.bar"
        )

    def _check_emails(self, actual, expected):
        actual_values = [(x.profile, x.address) for x in actual]
        expected_values = [(x.profile, x.address) for x in expected]
        self.assertEqual(actual_values, expected_values)

    def _check_accounts(self, actual, expected):
        act_values = [(x.profile, x.institution, x.cas_id) for x in actual]
        exp_values = [(x.profile, x.institution, x.cas_id) for x in expected]
        self.assertEqual(act_values, exp_values)

    @override_settings(UNIAUTH_PERFORM_RECURSIVE_MERGING=False)
    def test_merge_model_instances_non_recursive(self):
        """
        Ensure non-recursive merges work as expected
        """
        merge_model_instances(self.user1, [self.user2])
        self.assertTrue(
            User.objects.filter(username="user1@example.com").exists()
        )
        self.assertFalse(
            User.objects.filter(username="cas-test-uni-user2").exists()
        )

        merged = User.objects.get(username="user1@example.com")
        emails = list(merged.uniauth_profile.linked_emails.order_by("address"))
        accounts = list(merged.uniauth_profile.accounts.order_by("cas_id"))

        expected_emails = [
            LinkedEmail(
                profile=merged.uniauth_profile, address="user1@example.com"
            ),
        ]
        expected_accounts = []
        self._check_emails(emails, expected_emails)
        self._check_accounts(accounts, expected_accounts)

        merge_model_instances(self.user3, [self.user4, self.user5])
        self.assertTrue(
            User.objects.filter(username="user3@gmail.com").exists()
        )
        self.assertFalse(
            User.objects.filter(username="us.er4@nowhere.gov").exists()
        )
        self.assertFalse(
            User.objects.filter(username="user5@foo.bar").exists()
        )

        merged = User.objects.get(username="user3@gmail.com")
        emails = list(merged.uniauth_profile.linked_emails.order_by("address"))
        accounts = list(merged.uniauth_profile.accounts.order_by("cas_id"))

        expected_emails = [
            LinkedEmail(
                profile=merged.uniauth_profile, address="backup@gmail.com"
            ),
            LinkedEmail(
                profile=merged.uniauth_profile, address="user3@gmail.com"
            ),
        ]
        expected_accounts = [
            InstitutionAccount(
                profile=merged.uniauth_profile,
                institution=self.inst1,
                cas_id="john123",
            ),
        ]
        self._check_emails(emails, expected_emails)
        self._check_accounts(accounts, expected_accounts)

    @override_settings(UNIAUTH_PERFORM_RECURSIVE_MERGING=True)
    def test_merge_model_instances_recursive(self):
        """
        Ensure recursive merges work as expected
        """
        merge_model_instances(self.user1, [self.user2])
        self.assertTrue(
            User.objects.filter(username="user1@example.com").exists()
        )
        self.assertFalse(
            User.objects.filter(username="cas-test-uni-user2").exists()
        )

        merged = User.objects.get(username="user1@example.com")
        emails = list(merged.uniauth_profile.linked_emails.order_by("address"))
        accounts = list(merged.uniauth_profile.accounts.order_by("cas_id"))

        expected_emails = [
            LinkedEmail(
                profile=merged.uniauth_profile, address="user1@example.com"
            ),
        ]
        expected_accounts = [
            InstitutionAccount(
                profile=merged.uniauth_profile,
                institution=self.inst1,
                cas_id="user2",
            ),
        ]
        self._check_emails(emails, expected_emails)
        self._check_accounts(accounts, expected_accounts)

        merge_model_instances(self.user3, [self.user4, self.user5])
        self.assertTrue(
            User.objects.filter(username="user3@gmail.com").exists()
        )
        self.assertFalse(
            User.objects.filter(username="us.er4@nowhere.gov").exists()
        )
        self.assertFalse(
            User.objects.filter(username="user5@foo.bar").exists()
        )

        merged = User.objects.get(username="user3@gmail.com")
        emails = list(merged.uniauth_profile.linked_emails.order_by("address"))
        accounts = list(merged.uniauth_profile.accounts.order_by("cas_id"))

        expected_emails = [
            LinkedEmail(
                profile=merged.uniauth_profile, address="another@gmail.com"
            ),
            LinkedEmail(
                profile=merged.uniauth_profile, address="backup@gmail.com"
            ),
            LinkedEmail(
                profile=merged.uniauth_profile, address="huh@what.edu"
            ),
            LinkedEmail(
                profile=merged.uniauth_profile, address="user3@gmail.com"
            ),
            LinkedEmail(
                profile=merged.uniauth_profile, address="user5@foo.bar"
            ),
        ]
        expected_accounts = [
            InstitutionAccount(
                profile=merged.uniauth_profile,
                institution=self.inst2,
                cas_id="exid001",
            ),
            InstitutionAccount(
                profile=merged.uniauth_profile,
                institution=self.inst1,
                cas_id="john123",
            ),
        ]
        self._check_emails(emails, expected_emails)
        self._check_accounts(accounts, expected_accounts)



"""
This command is used to migrate a project previously using
CAS for authentication over to Uniauth.

Execution: python manage.py migrate_cas <slug>
"""

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from uniauth.models import Institution, InstitutionAccount, UserProfile
from uniauth.utils import get_input


class Command(BaseCommand):
    help = "Migrates a project using CAS to Uniauth."

    def add_arguments(self, parser):
        parser.add_argument("slug")

    @transaction.atomic
    def handle(self, *args, **options):
        slug = options["slug"]

        try:
            institution = Institution.objects.get(slug=slug)
        except Institution.DoesNotExist:
            raise CommandError("No institution with slug '%s' exists." % slug)

        message = (
            "This command is intended to migrate projects "
            "previously using CAS for authentication to using Uniauth.\n\nYou "
            "should only proceed with this command if your project was "
            "previously using CAS for authentication, and the usernames for "
            "all existing Users are equivalent to their CAS ID. This command "
            "will create UserProfiles for each user with the Institution "
            "specified by the slug argument.\n\nDo you still wish to "
            "continue?\n\nAnswer [y/n]: "
        )
        answer = get_input(message)

        if answer != "y" and answer != "yes":
            self.stdout.write("\nCanceled.\n")
            return

        self.stdout.write("\nProceeding... ")
        for user in get_user_model().objects.all():
            # Skip users that already have UserProfiles
            if hasattr(user, "uniauth_profile"):
                continue
            # Update the username to the proper format
            cas_id = user.username
            user.username = "cas-%s-%s" % (slug, cas_id)
            user.save()
            # Add the profile
            profile = UserProfile.objects.create(user=user)
        self.stdout.write("Done!\n")



import json

from django.contrib.auth.models import User
from django.core import mail
from django.http import HttpResponseRedirect
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse

from tests.utils import assert_urls_equivalent
from uniauth.models import Institution, InstitutionAccount, LinkedEmail
from uniauth.views import (
    _add_institution_account,
    _get_global_context,
    _login_success,
    _send_verification_email,
    get_jwt_tokens_from_session,
)


class AddInstitutionAccountTests(TestCase):
    """
    Tests the _add_instituion_account method in views.py
    """

    def setUp(self):
        self.inst1 = Institution.objects.create(
            name="Test Uni",
            slug="test-uni",
            cas_server_url="https://cas.testuni.edu",
        )
        self.inst2 = Institution.objects.create(
            name="Other Inst",
            slug="other-inst",
            cas_server_url="https://fed.other-inst.edu",
        )

    def test_add_institution_account_link_succeeds(self):
        """
        Ensures the method successfully links the new
        Institution Account to the profile
        """
        user = User.objects.create(username="johndoe")
        User.objects.create(username="cas-test-uni-jd123")
        _add_institution_account(user.uniauth_profile, "test-uni", "jd123")
        linked_accounts = user.uniauth_profile.accounts.all()
        cas_user1 = User.objects.get(username="cas-test-uni-jd123")
        self.assertEqual(len(linked_accounts), 1)
        self.assertEqual(linked_accounts[0].institution, self.inst1)
        self.assertEqual(linked_accounts[0].cas_id, "jd123")

        User.objects.create(username="cas-other-inst-john.doe4")
        _add_institution_account(
            user.uniauth_profile, "other-inst", "john.doe4"
        )
        linked_accounts = user.uniauth_profile.accounts.all()
        cas_user2 = User.objects.get(username="cas-other-inst-john.doe4")
        self.assertEqual(len(linked_accounts), 2)
        self.assertEqual(linked_accounts[1].institution, self.inst2)
        self.assertEqual(linked_accounts[1].cas_id, "john.doe4")


class GetGlobalContextTests(TestCase):
    """
    Tests the _get_global_context method in views.py
    """

    def setUp(self):
        self.factory = RequestFactory()
        Institution.objects.all().delete()
        self.inst1 = Institution.objects.create(
            name="Test Uni",
            slug="test-uni",
            cas_server_url="https://cas.testuni.edu",
        )
        self.inst2 = Institution.objects.create(
            name="Other Inst",
            slug="other-inst",
            cas_server_url="https://fed.other-inst.edu",
        )

    def test_get_global_context_correct(self):
        """
        Ensure method returns the expected template context
        """
        request = self.factory.get(
            "/accounts/login/", data={"foo": "bar", "next": "/next-page/"}
        )
        result = _get_global_context(request)
        result["institutions"].sort(key=lambda x: x[0])
        expected_institutions = [
            (
                "Other Inst",
                "other-inst",
                "/accounts/cas-login/other-inst/",
                "/accounts/link-from-profile/other-inst/",
            ),
            (
                "Test Uni",
                "test-uni",
                "/accounts/cas-login/test-uni/",
                "/accounts/link-from-profile/test-uni/",
            ),
        ]
        expected_params = "?foo=bar&next=%2Fnext-page%2F"
        self.assertEqual(
            sorted(list(result.keys())), ["institutions", "query_params"]
        )
        self.assertEqual(result["institutions"], expected_institutions)
        assert_urls_equivalent(
            result["query_params"], expected_params, self.assertEqual
        )


class LoginSuccessTests(TestCase):
    """
    Tests the _login_success method in views.py
    """

    factory = RequestFactory()

    def _run_test(self, username, params_dict, expected_url):
        user = User.objects.create(username=username)
        request = self.factory.get("/test/page/", data=params_dict)
        result = _login_success(request, user, "/next-page/")
        self.assertEqual(type(result), HttpResponseRedirect)
        assert_urls_equivalent(result.url, expected_url, self.assertEqual)

    def test_login_success_correct_destination(self):
        """
        Ensure method returns the correct destination URL,
        dependent on whether the user has a temporary account
        """
        self._run_test(
            "tmp-abc123", {}, "/accounts/link-to-profile/?next=%2Fnext-page%2F"
        )
        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=False):
            self._run_test(
                "cas-dashed-inst-id123",
                {},
                "/accounts/link-to-profile/?next=%2Fnext-page%2F",
            )
        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=True):
            self._run_test("cas-dashed-inst-id987", {}, "/next-page/")
        self._run_test("john.doe@hmail.moc", {}, "/next-page/")

    def test_login_success_preserve_query_params(self):
        """
        Ensure outputted URL preserves query parameters
        """
        self._run_test(
            "tmp-a0b1c2",
            {"hello": "world"},
            "/accounts/link-to-profile/?hello=world&next=%2Fnext-page%2F",
        )
        self._run_test(
            "tmp-z9y8x7",
            {"foo": "bar", "next": "/detour/"},
            "/accounts/link-to-profile/?foo=bar&next=%2Fnext-page%2F",
        )
        self._run_test("example@foo.bar", {"hi": "bye"}, "/next-page/?hi=bye")
        self._run_test(
            "student@inst.edu",
            {"dead": "beef", "next": "/detour/"},
            "/next-page/?dead=beef",
        )

    @override_settings(UNIAUTH_USE_JWT_AUTH=True)
    def test_login_success_use_jwt_auth(self):
        user = User.objects.create(username="student@institution.edu")

        self.client.force_login(user)
        session = self.client.session

        response = self.client.get(reverse("uniauth:login"))

        self.assertEqual(type(response), HttpResponseRedirect)
        self.assertTrue(session["jwt-refresh"])
        self.assertTrue(session["jwt-access"])


class SendVerificationEmail(TestCase):
    """
    Tests the _send_verification_email method in views.py
    """

    factory = RequestFactory()

    @override_settings(UNIAUTH_FROM_EMAIL="uniauth@testsmtp.ml")
    def test_send_verification_email_correct(self):
        """
        Ensure the subject, to address, and from address for
        a sent verification email is correct
        """
        user = User.objects.create(username="student")
        verify_email = LinkedEmail.objects.create(
            profile=user.uniauth_profile,
            address="toverify@gmail.com",
            is_verified=False,
        )
        request = self.factory.get("/accounts/somepage/", secure=False)
        _send_verification_email(request, "student@example.edu", verify_email)
        self.assertEqual(len(mail.outbox), 1)
        self.assertTrue("verify" in mail.outbox[0].subject.lower())
        self.assertTrue("/verify-token/" in mail.outbox[0].body)
        self.assertEqual(mail.outbox[0].to, ["student@example.edu"])
        self.assertEqual(mail.outbox[0].from_email, "uniauth@testsmtp.ml")

        user2 = User.objects.create(username="otherstud")
        verify_email2 = LinkedEmail.objects.create(
            profile=user2.uniauth_profile,
            address="stud_2@example.edu",
            is_verified=False,
        )
        request = self.factory.get("/accounts/otherpage/", secure=True)
        _send_verification_email(request, "stud_2@example.edu", verify_email2)
        self.assertEqual(len(mail.outbox), 2)
        self.assertTrue("verify" in mail.outbox[1].subject.lower())
        self.assertTrue("/verify-token/" in mail.outbox[1].body)
        self.assertEqual(mail.outbox[1].to, ["stud_2@example.edu"])
        self.assertEqual(mail.outbox[1].from_email, "uniauth@testsmtp.ml")


class GetJWTTokensFromSession(TestCase):
    """
    Tests the get_jwt_tokens_from_session method in views.py
    """

    factory = RequestFactory()

    FAKE_REFRESH_TOKEN = "refresh.token.string"
    FAKE_ACCESS_TOKEN = "access.token.string"

    def _run_test(
        self,
        username,
        session_state,
        expected_response_status,
        expected_response_data,
    ):
        user = User.objects.create(username=username)
        request = self.factory.get("/jwt-tokens/", data={})
        request.user = user
        request.session = session_state
        response = get_jwt_tokens_from_session(request)
        self.assertEqual(response.status_code, expected_response_status)
        self.assertEqual(
            json.loads(response.content.decode("utf-8")),
            expected_response_data,
        )

    def test_get_jwt_tokens_from_session(self):
        self._run_test(
            "tmp-c65433",
            {
                "jwt-refresh": self.FAKE_REFRESH_TOKEN,
                "jwt-access": self.FAKE_ACCESS_TOKEN,
            },
            200,
            {
                "refresh": self.FAKE_REFRESH_TOKEN,
                "access": self.FAKE_ACCESS_TOKEN,
            },
        )
        self._run_test(
            "tmp-a67653", {"jwt-refresh": self.FAKE_REFRESH_TOKEN}, 404, {}
        )
        self._run_test(
            "tmp-a65343", {"jwt-access": self.FAKE_ACCESS_TOKEN}, 404, {}
        )
        self._run_test("tmp-a75643", {}, 404, {})



"""
This command is used to add an Institution to the database.

Execution: python manage.py add_institution <name> <cas_server_url>
"""

from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand, CommandError
from django.core.validators import URLValidator
from django.utils.text import slugify

from uniauth.models import Institution


class Command(BaseCommand):
    help = "Adds an institution to the database."

    def add_arguments(self, parser):
        parser.add_argument("name")
        parser.add_argument("cas_server_url")
        parser.add_argument(
            "--update-existing",
            action="store_true",
            default=False,
            help="Update the institution, if it already exists.",
        )

    def handle(self, *args, **options):
        slug = slugify(options["name"])
        cas_server_url = options["cas_server_url"]

        if (
            not options["update_existing"]
            and Institution.objects.filter(slug=slug).exists()
        ):
            raise CommandError(
                "An institution with slug '" + slug + "' already exists."
            )

        try:
            validator = URLValidator()
            validator(options["cas_server_url"])
        except ValidationError:
            raise CommandError(
                "Provided CAS server URL '"
                + cas_server_url
                + "' is malformed."
            )

        institution, created = Institution.objects.get_or_create(
            name=options["name"],
            slug=slug,
            defaults={"cas_server_url": cas_server_url},
        )

        if created:
            self.stdout.write("Created institution '%s'.\n" % str(institution))
        elif institution.cas_server_url != cas_server_url:
            # If institution already exists but with a different URL,
            # update it.
            institution.cas_server_url = cas_server_url
            institution.save()
            self.stdout.write("Updated institution '%s'.\n" % str(institution))



"""
URL configuration for testing.
"""

try:
    from django.conf.urls import include, url
except ImportError:
    from django.urls import include
    from django.urls import re_path as url

urlpatterns = [
    url("accounts/", include("uniauth.urls", namespace="uniauth")),
]



"""
This command is used to flush old temporary accounts from the database.

Users with a username prefix of "tmp-" more than the specified number of
days old will be deleted. The default number of days is 1.

Execution: python manage.py flush_tmp_users [days]
"""

from django.core.management.base import BaseCommand, CommandError

from uniauth.utils import flush_old_tmp_users, get_input


class Command(BaseCommand):
    help = "Deletes old temporary accounts from the database."

    def add_arguments(self, parser):
        parser.add_argument("days", type=int, nargs="?", default=1)

    def handle(self, *args, **options):
        days = options["days"]

        answer = get_input(
            "Are you sure you want to delete all temporary"
            + "users more than %d days old?\nAnswer [y/n]:" % days
        )
        if answer == "y" or answer == "yes":
            num_deleted = flush_old_tmp_users(days=days)
            self.stdout.write("Deleted %d temporary users.\n")
        else:
            self.stdout.write("Canceled.\n")



# Generated by Django 2.1.4 on 2019-01-25 17:51

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Institution",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=30)),
                ("slug", models.CharField(max_length=30, unique=True)),
                ("cas_server_url", models.URLField()),
            ],
        ),
        migrations.CreateModel(
            name="InstitutionAccount",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("cas_id", models.CharField(max_length=30)),
                (
                    "institution",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="accounts",
                        to="uniauth.Institution",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="LinkedEmail",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("address", models.EmailField(max_length=254)),
                ("is_verified", models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name="UserProfile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="profile",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="linkedemail",
            name="profile",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="linked_emails",
                to="uniauth.UserProfile",
            ),
        ),
        migrations.AddField(
            model_name="institutionaccount",
            name="profile",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="accounts",
                to="uniauth.UserProfile",
            ),
        ),
    ]
